"""Sample Sphinx config file for use with behave-based testing repositories."""

import os


# PLEASE EDIT THE FOLLOWING CONFIGURATION INFORMATION:

# General information about your project.
project = "%%PROJECT%%"
copyright = "2018, Rackspace Quality Engineering"  # noqa
author = "%%AUTHOR%%"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = "%%VERSION%%"

# The full version, including alpha/beta/rc tags.
release = "%%RELEASE%%"

# BELOW HERE YOU SHOULD BE ABLE TO LEAVE AS-IS.


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

from recommonmark.parser import CommonMarkParser  # noqa

source_parsers = {".md": CommonMarkParser}
source_suffix = [".rst", ".md"]

# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    ".tox",
    "*/.tox",
    ".eggs",
    "*/.eggs",
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

# If anyone wants to use another theme, they can change that here,
# but we consider that expert Sphinx user territory.
import sphinx_rtd_theme  # noqa

html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

commit_id = os.environ.get("ghprbPullId") or os.environ.get("GIT_COMMIT_ID")
base_url = os.environ.get("ghprbPullLink")
if not base_url:
    owner_name = os.environ.get("GIT_ORIGIN_URL", "")
    if not owner_name:
        base_url = ""  # If this is non-empty, sphinx will make it clickable.
    else:
        owner_name = os.path.splitext(owner_name.split(":")[1])[0]
        base_url = "https://github.rackspace.com/{}/tree/{}".format(
            owner_name, commit_id
        )
html_context = {"build_id": commit_id, "build_url": base_url}



"""File-related utils used throughout the module."""
import fnmatch
import os
import pathlib
from typing import Iterable, List, NamedTuple


# Must be lowercase.
FEATURE_FILE_SUFFIX = ".feature"
SOURCE_SUFFICIES = (FEATURE_FILE_SUFFIX, ".md", ".rst")


class DirData(NamedTuple):
    """Named tuple containing all the modified os.walk parts for a given path."""

    dir_path: pathlib.Path
    path_list: List[str]
    sub_dirs: List[str]
    files: List[str]


def is_feature_file(filename: str) -> bool:
    """Determine if the given filename is a feature file."""
    return filename.lower().endswith(FEATURE_FILE_SUFFIX)


def is_rst_file(filename: str) -> bool:
    """Determine if the given filename is a rST file."""
    return filename.lower().endswith(".rst")


def _not_private(filename: str) -> bool:
    """Check if a filename is private (using underscore prefix)."""
    return not filename.startswith("_")


def _is_wanted_file(filename: str) -> bool:
    """Wanted as in: We know how to process it."""
    return filename.lower().endswith(SOURCE_SUFFICIES)


def _is_excluded(filename: str, exclude_pattern_list: Iterable) -> bool:
    """Determine if the given filename should be excluded, based on the pattern list."""
    return any(map(lambda x: fnmatch.fnmatch(filename, x), exclude_pattern_list))


def _not_hidden(name: str) -> bool:
    """Determine if the filename is not hidden."""
    return not name.startswith(".")


def _wanted_source_files(files: Iterable, exclude_pattern_list: Iterable) -> List[str]:
    """Get list of wanted sorce files, excluding unwanted files."""
    wanted_files = filter(_is_wanted_file, files)
    return [
        a_file
        for a_file in wanted_files
        if not _is_excluded(a_file, exclude_pattern_list)
    ]


def scan_tree(
    starting_point: pathlib.Path, private: bool, exclude_patterns: Iterable
) -> List[DirData]:
    """
    Return list of entities to process, in top-down orders.

    the list can easily be processed with `.pop` to get a bottom-up
    directory ordering.
    """
    result = []

    # Do not assume an absolute path is provided
    starting_point = starting_point.resolve()
    for me, dirs, files in os.walk(starting_point):
        if _is_excluded(me, exclude_patterns):
            dirs[:] = []
            continue

        # Remove all hidden directories on principle.
        # This stops scanning into version control directories such as .git, ,hg, etc.
        dirs[:] = filter(_not_hidden, dirs)

        if not private:
            dirs[:] = filter(_not_private, dirs)

        me_path = pathlib.Path(me)
        result.append(
            DirData(
                me_path,
                [str(x) for x in me_path.relative_to(starting_point.parent).parts],
                dirs[:],
                _wanted_source_files(files, exclude_patterns),
            )
        )

    return result



"""Module specific to building the steps glossary."""
from collections import defaultdict
import pathlib
from typing import DefaultDict, Optional, Set, Tuple

from .utils import INDENT_DEPTH, rst_escape, SphinxWriter


class GlossaryEntry(object):
    """Track all the different places, and by which spellings, something is used."""

    def __init__(self) -> None:
        self.step_set: Set[str] = set()
        self.locations: DefaultDict[pathlib.Path, list] = defaultdict(list)

    def add_reference(
        self, step_name: str, filename: pathlib.Path, line_number: int
    ) -> None:
        """
        Add a reference to the glossary.

        Args:
            step_name: The step name in Gherkin
            filename: The file name containing the step
            line_number: The line number containing the step

        """

        self.step_set.add(step_name)
        self.locations[filename].append(line_number)

    def tuple_len(self) -> Tuple[int, int]:
        """Get the length for each location and the number of associated steps."""
        return (len(self.locations), sum(map(len, self.locations.values())))

    def __gt__(self, other) -> bool:
        """Compare the location and step lenghts for the larger one."""
        return self.tuple_len() > other.tuple_len()

    def __eq__(self, other) -> bool:
        """Compare the location and step lengths for equality."""
        return self.tuple_len() == other.tuple_len()


step_glossary: DefaultDict[str, GlossaryEntry] = defaultdict(GlossaryEntry)


def make_steps_glossary(project_name: str) -> Optional[SphinxWriter]:
    """Return SphinxWriter containing the step glossary information, if any."""

    if not step_glossary:
        return None

    glossary = SphinxWriter()
    glossary.create_section(1, "{} Glossary".format(project_name))

    master_step_names = {
        name for gloss in step_glossary.values() for name in gloss.step_set
    }
    for term in sorted(master_step_names):
        glossary.add_output("- :term:`{}`".format(rst_escape(term, slash_escape=True)))

    glossary.blank_line()
    glossary.add_output(".. glossary::")
    for entry in sorted(step_glossary.values(), reverse=True):
        for term in sorted(entry.step_set):
            glossary.add_output(
                rst_escape(term, slash_escape=True), indent_by=INDENT_DEPTH
            )

        for location, line_numbers in sorted(entry.locations.items()):
            definition = f"| {location} {', '.join(map(str, line_numbers))}"
            glossary.add_output(definition, indent_by=INDENT_DEPTH * 2)
        glossary.blank_line()

    return glossary



#!/usr/bin/env python3
"""Module for running the tool from the CLI."""
import argparse
import pathlib
import shutil
from typing import Set

import sphinx

from .files import is_feature_file, is_rst_file, scan_tree
from .glossary import make_steps_glossary
from .utils import make_flat_name, set_dry_run, set_verbose, verbose
from .writer import feature_to_rst, toctree

# This is a pretty arbitrary number controlling how much detail
# will show up in the various TOCs.
DEFAULT_TOC_DEPTH = 4


def process_args(
    args: argparse.Namespace,
    gherkin_path: pathlib.Path,
    output_path: pathlib.Path,
    doc_project: str,
) -> None:
    """Process the supplied CLI args."""
    work_to_do = scan_tree(gherkin_path, args.private, args.exclude_patterns)
    maxtocdepth = args.maxtocdepth
    toc_name = args.toc_name
    step_glossary_name = args.step_glossary_name
    doc_project = args.doc_project
    root_path = gherkin_path.resolve().parent

    top_level_toc_filename = output_path / f"{toc_name}.rst"

    non_empty_dirs: Set[pathlib.Path] = set()

    while work_to_do:
        current = work_to_do.pop()
        new_subdirs = []
        for subdir in current.sub_dirs:
            subdir_path = pathlib.Path() / current.dir_path / subdir
            if subdir_path in non_empty_dirs:
                new_subdirs.append(subdir)

        if not (current.files or new_subdirs):
            continue

        non_empty_dirs.add(current.dir_path)

        if args.dry_run:
            continue

        toc_file = toctree(
            current.path_list, new_subdirs, current.files, maxtocdepth, root_path
        )
        # Check to see if we are at the last item to be processed
        # (which has already been popped)
        # to write the asked for master TOC file name.
        if not work_to_do:
            toc_filename = top_level_toc_filename
        else:
            toc_filename = output_path / make_flat_name(current.path_list, is_dir=True)
        toc_file.write_to_file(toc_filename)

        for a_file in current.files:
            a_file_list = current.path_list + [a_file]
            source_name = pathlib.Path().joinpath(*a_file_list)
            source_path = root_path / source_name
            if is_feature_file(a_file):
                dest_name = output_path / make_flat_name(a_file_list, is_dir=False)
                feature_rst_file = feature_to_rst(
                    source_path, root_path, args.url_from_tag
                )
                verbose('converting "{}" to "{}"'.format(source_name, dest_name))
                feature_rst_file.write_to_file(dest_name)
            elif not is_rst_file(a_file):
                dest_name = output_path / make_flat_name(
                    a_file_list, is_dir=False, ext=None
                )
                verbose('copying "{}" to "{}"'.format(source_name, dest_name))
                shutil.copy(source_path, dest_name)

    if step_glossary_name:
        glossary_filename = output_path / f"{step_glossary_name}.rst"
        glossary = make_steps_glossary(doc_project)

        if args.dry_run:
            verbose("No glossary generated")
            return

        if glossary is None:
            print("No steps to include in the glossary: no glossary generated")
            return

        verbose("Writing sphinx glossary: {}".format(glossary_filename))
        glossary.write_to_file(glossary_filename)


def main() -> None:
    """Convert a directory-tree of Gherkin Feature files to rST files."""
    description = (
        "Look recursively in <gherkin_path> for Gherkin files and create one "
        "reST file for each. Other rST files found along the way will be included "
        "as prologue content above each TOC."
    )
    parser = argparse.ArgumentParser(
        description=description, formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("gherkin_path", help="Directory to search for Gherkin files")
    parser.add_argument("output_path", help="Directory to place all output")
    parser.add_argument(
        "exclude_patterns",
        nargs="*",
        help="file and/or directory patterns that will be excluded",
    )
    parser.add_argument(
        "-d",
        "--maxtocdepth",
        type=int,
        default=DEFAULT_TOC_DEPTH,
        help="Maximum depth of submodules to show in the TOC",
    )
    parser.add_argument(
        "-n",
        "--dry-run",
        action="store_true",
        help="Run the script without creating files",
    )
    parser.add_argument(
        "-P", "--private", action="store_true", help='Include "_private" folders'
    )
    parser.add_argument("-N", "--toc-name", help="File name for TOC", default="gherkin")
    parser.add_argument(
        "-H", "--doc-project", help="Project name (default: root module name)"
    )
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="Silence any output to screen"
    )
    parser.add_argument(
        "-G",
        "--step-glossary-name",
        default=None,
        help="Include steps glossary under the given name."
        " If not specified, no glossary will be created.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="print files created and actions taken",
    )
    parser.add_argument(
        "--version", action="store_true", help="Show version information and exit"
    )
    url_help = (
        "A library and method name to call to build a URL from a tag. The string"
        " should be <library>:<method_name> and it should accept a single string"
        " parameter, the tag."
    )
    parser.add_argument("--url-from-tag", help=url_help)

    args = parser.parse_args()

    if args.version:
        parser.exit(
            message="Sphinx (sphinx-gherkindoc) {}".format(sphinx.__display_version__)
        )

    set_dry_run(args.dry_run)
    set_verbose(args.verbose)

    gherkin_path = pathlib.Path(args.gherkin_path)
    if not gherkin_path.is_dir():
        parser.error("{} is not a directory.".format(gherkin_path))

    if args.doc_project is None:
        args.doc_project = gherkin_path.parts[-1]

    output_path = pathlib.Path(args.output_path).resolve()
    if not output_path.is_dir():
        if not args.dry_run:
            verbose("creating directory: {}".format(output_path))
            output_path.mkdir(parents=True)

    process_args(args, gherkin_path, output_path, args.doc_project)


def config() -> None:
    """Emit a customized version of the sample sphinx config file."""
    description = (
        "Create a default Sphinx configuration for producing nice"
        " Gherkin-based documentation"
    )
    parser = argparse.ArgumentParser(
        description=description, formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "project_name", default="Your Project Name Here", help="Name of your project"
    )
    parser.add_argument(
        "author", default="Your Team Name Here", help="Directory to place all output"
    )
    parser.add_argument("--version", default="", help="version of your project, if any")
    parser.add_argument("--release", default="", help="release of your project, if any")
    args = parser.parse_args()

    substitutions = {
        "%%PROJECT%%": args.project_name,
        "%%AUTHOR%%": args.author,
        "%%VERSION%%": args.version,
        "%%RELEASE%%": args.release,
    }
    source_dir = pathlib.Path(__file__).resolve().parent
    with open(source_dir / "sample-conf.py", "r") as conf_fo:
        sample_contents = conf_fo.read()
    for old_value, new_value in substitutions.items():
        sample_contents = sample_contents.replace(old_value, new_value)

    print(sample_contents)


if __name__ == "__main__":
    main()



"""Helper functions for writing rST files."""
import importlib
import itertools
import pathlib
import re
import pkg_resources
from typing import List, Optional, Union

import behave.parser
import behave.model
import behave.model_core

from .files import is_rst_file
from .glossary import step_glossary
from .utils import (
    display_name,
    make_flat_name,
    INDENT_DEPTH,
    rst_escape,
    SphinxWriter,
    verbose,
)

# The csv-table parser for restructuredtext does not allow for escaping so use
# a unicode character that looks like a quote but will not be in any Gherkin
QUOTE = "\u201C"


# IMPLEMENTATION NOTES:
# When generating rST output files from directory trees containing feature files,
# and rST/MarkDown, files, implementation concerns:
# * Walking a directory tree top down makes it hard to know if any subdirectories
#   will have interesting content worth including.
# * Walking a directory tree bottom up makes it hard to efficiently process excluded
#   directories.
#
# Strategy: Use a 2-phased approach:
# 1 Walk the directory tree top-down to avoid descending in to excluded directories.
#   Record an ordered list of places that are not excluded and thus worth looking at.
# 2 Process the recorded list in reverse order, effecting a bottom-up processing of the
#   tree, without the bottom-up concern. This way we can prune out directories that are
#   either empty or have empty children.


def toctree(
    path_list: List[str],
    subdirs: List[str],
    files: List[str],
    maxtocdepth: int,
    root_path: pathlib.Path,
) -> SphinxWriter:
    """
    Return a SphinxWriter for one level of a directory tree.

    Any rST files found at this level will have their content
    included in the TOC file instead of being referenced.
    This allows us to put content in the TOC file directly.
    The rST files that are included are expected to contain proper headers as well as
    content.
    If no rST files are included, a header is created using display_name.

    NOTE: If there is more than one rST file present,
          the order of their inclusion in the TOC is by filename sort order.
    """

    of = SphinxWriter()
    non_included_files = []
    need_header = True  # Track if we need to generate our own header boiler-plate.

    for a_file in sorted(files):
        if not is_rst_file(a_file):
            non_included_files.append(a_file)
            continue

        source_name_list = path_list + [a_file]
        source_name = pathlib.Path().joinpath(*source_name_list)
        source_path = root_path / source_name
        verbose("Copying content from: {}".format(source_name))
        with open(source_path, "r") as source_fo:
            of.add_output(source_fo.read(), line_breaks=2)
        need_header = False

    if need_header:
        # We're just adding a boiler plate heading.
        of.create_section(1, display_name(root_path.joinpath(*path_list)))

    of.add_output(".. toctree::")
    of.add_output(
        ":maxdepth: {}".format(maxtocdepth), line_breaks=2, indent_by=INDENT_DEPTH
    )

    for a_file in sorted(non_included_files):
        # For MarkDown file content to be properly processed we
        # need to leave the file name extension unchanged.
        ext = None if a_file.lower().endswith(".md") else ""
        name = make_flat_name(path_list, filename_root=a_file, is_dir=False, ext=ext)
        of.add_output(name, indent_by=INDENT_DEPTH)

    for subdir in sorted(subdirs):
        name = make_flat_name(path_list, filename_root=subdir, is_dir=True, ext="")
        of.add_output(name, indent_by=INDENT_DEPTH)

    return of


# Simplified this from a class, for various reasons.
# Additional simplification work is needed!!!!
def feature_to_rst(
    source_path: pathlib.Path, root_path: pathlib.Path, url_from_tag: Optional[str] = ""
) -> SphinxWriter:
    """Return a SphinxWriter containing the rST for the given feature file."""
    output_file = SphinxWriter()

    def section(level: int, obj: behave.model_core.BasicStatement) -> None:
        section_name = u"{}: {}".format(obj.keyword, rst_escape(obj.name))
        output_file.create_section(level, section_name.rstrip(": "))

    def description(description: Union[str, List[str]]) -> None:
        if not description:
            return
        if not isinstance(description, list):
            description = [description]
        for line in description:
            output_file.add_output(rst_escape(line), indent_by=INDENT_DEPTH)
            # Since behave strips newlines, a reasonable guess must be made as
            # to when a newline should be re-inserted
            if line[-1] == "." or line == description[-1]:
                output_file.blank_line()

    def text(text: Union[str, List[str]]) -> None:
        if not text:
            return
        if not isinstance(text, list):
            text = [text]
        output_file.blank_line()
        output_file.add_output("::", line_breaks=2)
        # Since text blocks are treated as raw text, any new lines in the
        # Gherkin are preserved. To convert the text block into a code block,
        # each new line must be indented.
        for line in itertools.chain(*(x.splitlines() for x in text)):
            output_file.add_output(
                rst_escape(line), line_breaks=2, indent_by=INDENT_DEPTH
            )

    # Build the URL parser once since `ticket_url_or_tag` is called multiple times
    url_parser = lambda x: ""  # noqa: E731
    for entry_point in pkg_resources.iter_entry_points("parsers"):
        if entry_point.name == "url":
            url_parser = entry_point.load()
    if url_from_tag:
        url_module, url_function = url_from_tag.split(":", maxsplit=1)
        parser_module = importlib.import_module(url_module)
        url_parser = getattr(parser_module, url_function)

    # Reference link here because it's too long to put inside the function itself.
    # http://docutils.sourceforge.net/docs/ref/rst/restructuredtext.html#embedded-uris-and-aliases
    def ticket_url_or_tag(tag: str) -> str:
        """Get a URL or tag.

        If tag is a ticket, return an anonymous embedded hyperlink for it,
        else tag itself.
        """
        url = url_parser(tag)
        if url:
            return "`{} <{}>`__".format(tag, url)
        return tag

    def tags(tags: List[str], *parent_objs: behave.model_core.BasicStatement) -> None:
        parent_with_tags = tuple(x for x in parent_objs if x.tags)
        if not (tags or parent_with_tags):
            return
        output_file.add_output(".. pull-quote::", line_breaks=2)
        tag_str = ", ".join(map(ticket_url_or_tag, tags))
        for obj in parent_with_tags:
            tags_list = ", ".join(map(ticket_url_or_tag, obj.tags))
            tag_str += u" (Inherited from {}: {} )".format(obj.keyword, tags_list)
        output_file.add_output(
            u"Tagged: {}".format(tag_str.strip()), line_breaks=2, indent_by=INDENT_DEPTH
        )

    def steps(steps: List[behave.model.Step]) -> None:
        for step in steps:
            step_glossary[step.name.lower()].add_reference(
                step.name,
                pathlib.Path(step.filename).resolve().relative_to(root_path),
                step.line,
            )
            bold_step = re.sub(r"(\\\<.*?\>)", r"**\1**", rst_escape(step.name))
            output_file.add_output(u"- {} {}".format(step.keyword, bold_step))
            if step.table:
                output_file.blank_line()
                table(step.table, inline=True)
            if step.text:
                text(step.text)
        output_file.blank_line()

    def examples(
        scenario: behave.model.Scenario, feature: behave.model.Feature
    ) -> None:
        for example in getattr(scenario, "examples", []):
            section(3, example)
            tags(example.tags, scenario, feature)
            table(example.table)
            output_file.blank_line()

    def table(table: behave.model.Table, inline: bool = False) -> None:
        indent_by = INDENT_DEPTH if inline else 0
        directive = ".. csv-table::"
        output_file.add_output(directive, indent_by=indent_by)
        headers = u'", "'.join(table.headings)
        indent_by += INDENT_DEPTH
        output_file.add_output(u':header: "{}"'.format(headers), indent_by=indent_by)
        output_file.add_output(
            u":quote: {}".format(QUOTE), line_breaks=2, indent_by=indent_by
        )
        for row in table.rows:
            row = u"{0}, {0}".format(QUOTE).join(map(rst_escape, row))
            output_file.add_output(u"{0}{1}{0}".format(QUOTE, row), indent_by=indent_by)

    feature = behave.parser.parse_file(source_path)
    section(1, feature)
    description(feature.description)
    if feature.background:
        section(2, feature.background)
        steps(feature.background.steps)
    for scenario in feature.scenarios:
        section(2, scenario)
        tags(scenario.tags, feature)
        description(scenario.description)
        steps(scenario.steps)
        examples(scenario, feature)

    return output_file



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['sphinx_gherkindoc']

package_data = \
{'': ['*']}

install_requires = \
['Sphinx>=1.3',
 'behave>=1.2.6',
 'recommonmark>=0.4.0',
 'sphinx_rtd_theme>=0.3.1']

entry_points = \
{'console_scripts': ['sphinx-gherkinconfig = sphinx_gherkindoc.cli:config',
                     'sphinx-gherkindoc = sphinx_gherkindoc.cli:main']}

setup_kwargs = {
    'name': 'sphinx-gherkindoc',
    'version': '3.2.1',
    'description': 'A tool to convert Gherkin into Sphinx documentation',
    'long_description': 'sphinx-gherkindoc\n=================\n\n``sphinx-gherkindoc`` brings Gherkin goodness\ninto the Sphinx/reStructuredText (rST) world.\n\nWhy should I use it?\n--------------------\n\n**Share your requirements with your larger team.**\nGherkin makes it easy for anyone to read requirements.\nSphinx makes Gherkin pretty enough to be readable.\n``sphinx-gherkindoc`` handles converting flat-text feature files\ninto easy to read documents you won\'t cringe to share with your larger team.\n\n**Easily see what steps you have and where you are using them.**\n``sphinx-gherkindoc`` can create a glossary of your steps.\nThis helps makes it easy to:\n\n   * find steps to reuse\n   * notice similar steps that might be duplicates or unintended variations\n   * see patterns that you might exploit to reduce the number of steps you have\n   * find out where and which feature files would be affected\n     if a step\'s wording or implementation changes\n\n\nHow does it do that?\n--------------------\n\n``sphinx-gherkindoc`` recursively scans a given directory\nto find all the feature and markup files,\nand converts them into files\nthat can be included in Sphinx-based documentation.\nThis script was inspired by ``sphinx-apidoc``\nwhich does a similar thing for source code.\n\n.. Note::\n\n    This tool only creates the rST inputs for a sphinx document run,\n    you still need to fit these files\n    in to your larger documentation building process yourself.\n\nFor specific details on the command line options,\nplease consult the ``--help`` output.\nMost command line options mirror their counterpart in ``sphinx-apidoc``.\n\nFor the most basic usage, an input (``<gherkin_path>``)\nand output (``<output_path>``) path must be provided.\nThe files put in to the ``<output_path>``\ncan be incorporated into any larger documentation,\ncreated by any means.\n\nAdditionally, a list of ``fnmatch``-compatible patterns can be added\nto the command line,\nto indicate directories to be excluded from processing.\n\nOne notable addition is the step glossary(``-G``, ``--step-glossary``).\nThe step glossary command line option causes ``sphinx-gherkindoc``\nto create the named file in the ``<output_path>`` directory.\nThe step glossary content is formatted into two lists:\n\n   * A list of all the steps found, in alphabetical order.\n     Each item in this list is a link to its details in the second list.\n   * A list of the steps showing the file names and line numbers\n     where they are used.\n     This list is in order by the most number of uses first.\n\n\nHow are my files converted?\n---------------------------\n\nWhen scanning the ``<gherkin_path>`` directory tree,\n``sphinx-gherkindoc`` will do the following:\n\n   * Feature files found\n     are processed into rST files in the ``<output_path>`` directory.\n   * Directories will be converted to Sphinx Table of Contents (TOC) files that\n     link to any feature files in the same directory,\n     and to any TOC files from direct subdirectories.\n   * Any rST files found in a directory have their contents copied\n     to the front of the TOC file for that directory.\n     If more than one rST file is found in a directory,\n     they are processed in sort order.\n   * If no rST files are found in a directory,\n     then a header for the TOC is created based\n     on the contents of a ``display_name.txt`` file, if present,\n     or the name of the directory.\n   * Any MarkDown (md) files are referenced\n     from the TOC file for the directory they are in.\n   * Directories with no feature, rST, or md files are pruned,\n     recursively upwards.\n\n\nThe meat\'n\'potatoes will be your feature files.\nPut rST files next to your feature files\nto present context and additional helpful material.\nIf any rST files are in the same directory,\nthey should also contain any appropriate headers\nand other such formatting.\n``sphinx-gherkindoc`` will only create (minimal) headers when\nthere are no rST files present at all.\n\n\nExamples\n--------\n\nDisclaimer: This `is not` a tutorial on how to use or configure Sphinx.\nIt `is` some common ways you can use ``sphinx-gherkindoc``\nas part of your documentation generation.\nReminder if you skipped all that stuff above:\n``--help`` will show you the default values\nwhen command line options aren\'t used.\n\nConventions - Sphinx-based document production usually uses two directories:\n\n    * ``_docs`` - the working directory where we put all the rST files\n      from the various tools as we are building documents.\n      This directory shoud not be checked in to version control\n      and should only contain files created by a documentation run.\n    * ``docs`` - the output directory for a documentation run.\n      For example, this is where ``index.html`` is found\n      when building HTML docs, etc.\n      This directory should not be checked in to version control\n      as it contains only derived and processed files.\n\n\n1. Convert feature files to rST;\n   process all feature and document files\n   in/under ``feature/root/directory-here`` into the ``_docs`` directory\n   using all the defaults::\n\n       sphinx-gherkindoc feature/root/directory-here _docs\n\n2. Same as above,\n   but also create a step glossary file ``my_step_glossary`` in ``_docs``::\n\n       sphinx-gherkindoc -G my_step_glossary feature/root/directory-here _docs\n\n3. Experiment!\nOnce you have the 2nd step working\nand integrated in to your document building process,\nyou may find you want to tweak the results some.\nIt\'s a lot easier to do that `after` you have the basic process working.\nExperiment with the other optional parameters\nto get the effect(s) you want.\n\nSpecial Instructions for Tag Links\n----------------------------------\n\nAny tag can be converted into an anonymous link\nvia a plugin or command-line argument.\nThe converter needs to be a single function\nthat accepts a single string parameter, the tag,\nand returns a URL if the tag should include a link\nor an empty string if not.\n\nIn order to register the plugin for a ``poetry``-based project::\n\n    [tools.poetry.plugins."parsers"]\n    url = "my_custom_library.parse:optional_url_from_tag"\n\nIn order to register the plugin for a ``setup.py``-based project::\n\n    setup(\n        ...\n        entry_points={\n            \'parsers\': [\'url = my_custom_library.parse:optional_url_from_tag\']\n        }\n    )\n\nIn order to use the parser via command line,\nthe ``--url_from-tag`` flag should be used.\nThe provided string should be be formatted ``<library>:<method_name>``\n',
    'author': 'Lewis Franklin',
    'author_email': 'lewis.franklin@gmail.com',
    'url': 'https://jolly-good-toolbelt.github.io/sphinx_gherkindoc/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Generic utils used throughout the module."""
import pathlib
import string
from typing import List, Optional

import sphinx.util

# Increments of how much we indent Sphinx rST content when indenting.
INDENT_DEPTH = 4


# DRY_RUN and VERBOSE are global states for all the code.
# By making these into global variables, the code "admits that" they are global;
# rather than cluttering up method parameters passing these values around,
# and having to track if any particular method/function needs or no-longer needs them.
DRY_RUN = False
VERBOSE = False


def verbose(message: str) -> None:
    """Print message only if VERBOSE, with a DRY_RUN prefix as appropriate."""
    if not VERBOSE:
        return
    if DRY_RUN:
        message = "dry-run: " + message
    print(message)


def set_dry_run(value: bool) -> None:
    """Set the value for DRY_RUN outside this module."""
    global DRY_RUN
    DRY_RUN = value


def set_verbose(value: bool) -> None:
    """Set the value for VERBOSE outside this module."""
    global VERBOSE
    VERBOSE = value


# Build up dictionary of characters that need escaping
_escape_mappings = {ord(x): u"\\{}".format(x) for x in ("*", '"', "#", ":", "<", ">")}
_advanced_escape_mappings = _escape_mappings.copy()
_advanced_escape_mappings[ord("\\")] = u"\\\\\\"


def rst_escape(unescaped: str, slash_escape: bool = False) -> str:
    """
    Escape reST-ful characters to prevent parsing errors.

    Args:
        unescaped: A string that potentially contains characters needing escaping
        slash_escape: if True, escape slashes found in ``unescaped``

    Returns:
        A string which has reST-ful characters appropriately escaped

    """
    return unescaped.translate(
        _advanced_escape_mappings if slash_escape else _escape_mappings
    )


def make_flat_name(
    path_list: List[str],
    filename_root: str = None,
    is_dir: bool = False,
    ext: Optional[str] = ".rst",
) -> str:
    """
    Build a flat file name from the provided information.

    Args:
        path_list: Directory hierarchy to flatten
        filename_root: If provided, the root of the filename to flatten (no extension)
        is_dir: If True, mark the new filename as a table of contents
        ext: Optional extension for the new file name

    Returns:
        A filename containing the full path, separated by periods

    """
    if filename_root is not None:
        path_list = path_list + [filename_root]
    result = ".".join(path_list)
    if ext is None:
        return result
    return result + ("-toc" if is_dir else "-file") + ext


class SphinxWriter(object):
    """Easy Sphinx-format file creator."""

    sections = ["", "=", "-", "~", ".", "*", "+", "_", "<", ">", "/"]

    def __init__(self) -> None:
        self._output: List[str] = []

    def add_output(self, line: str, line_breaks: int = 1, indent_by: int = 0) -> None:
        """Add output to be written to file.

        Args:
            line: The line to be written
            line_breaks: The number of line breaks to include
            indenty_by: The number of spaces to indent the line.
        """
        self._output.append(u"{}{}{}".format(" " * indent_by, line, "\n" * line_breaks))

    def blank_line(self) -> None:
        """Write a single blank line."""
        self.add_output("")

    def create_section(self, level: int, section: str) -> None:
        """
        Create a reST-formatted section header based on the provided level.

        Args:
            level: The level depth of the section header (1-10 supported)
            section: The section title
        """
        self.add_output(section)
        self.add_output(self.sections[level] * len(section.rstrip()), line_breaks=2)

    def write_to_file(self, filename: pathlib.Path) -> None:
        """Write the provided output to the given filename.

        Args:
            filename: The full path to write the output
        """
        verbose("Writing {}".format(filename))
        with sphinx.util.osutil.FileAvoidWrite(filename) as f:
            # All version of Sphinx will accept a string-type,
            # but >=2.0 accepts _only_ strings (not bytes)
            data = "".join(self._output)
            if not isinstance(data, str):
                # py3 will recognize unicode as strings natively,
                # py2 unicode needs to be encoded first
                data = data.encode("utf8")
            f.write(data)


def display_name(path: pathlib.Path, package_name: Optional[str] = "") -> str:
    """
    Create a human-readable name for a given project.

    Determine the display name for a project given a path and (optional) package name.
    If a display_name.txt file is found, the first line is returned. Otherwise, return a
    title-cased string from either the base directory or package_name (if provided).

    Args:
        path: Path for searching
        package_name: Sphinx-style, dot-delimited package name (optional)

    Returns:
        A display name for the provided path

    """
    name_path = path / "display_name.txt"
    if name_path.exists():
        with open(name_path, "r") as name_fo:
            return name_fo.readline().rstrip("\r\n")
    raw_name = package_name.split(".")[-1] if package_name else path.name
    return string.capwords(raw_name.replace("_", " "))
